# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'GlacierInventory'
        db.create_table('icybackup_glacierinventory', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('inventory_id', self.gf('django.db.models.fields.CharField')(unique=True, max_length=92)),
            ('collected_date', self.gf('django.db.models.fields.DateTimeField')(default=None, null=True, blank=True)),
        ))
        db.send_create_signal('icybackup', ['GlacierInventory'])


    def backwards(self, orm):
        # Deleting model 'GlacierInventory'
        db.delete_table('icybackup_glacierinventory')


    models = {
        'icybackup.glacierbackup': {
            'Meta': {'object_name': 'GlacierBackup'},
            'date': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'glacier_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '138'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'})
        },
        'icybackup.glacierinventory': {
            'Meta': {'object_name': 'GlacierInventory'},
            'collected_date': ('django.db.models.fields.DateTimeField', [], {'default': 'None', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'inventory_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '92'})
        }
    }

    complete_apps = ['icybackup']


# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):

        # Changing field 'GlacierBackup.date'
        db.alter_column('icybackup_glacierbackup', 'date', self.gf('django.db.models.fields.DateTimeField')())

    def backwards(self, orm):

        # Changing field 'GlacierBackup.date'
        db.alter_column('icybackup_glacierbackup', 'date', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True))

    models = {
        'icybackup.glacierbackup': {
            'Meta': {'object_name': 'GlacierBackup'},
            'date': ('django.db.models.fields.DateTimeField', [], {}),
            'glacier_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '138'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'})
        },
        'icybackup.glacierinventory': {
            'Meta': {'object_name': 'GlacierInventory'},
            'collected_date': ('django.db.models.fields.DateTimeField', [], {'default': 'None', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'inventory_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '92'}),
            'requested_date': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['icybackup']


from .. import models
from datetime import timedelta, datetime
from boto.glacier.layer2 import Layer2 as Glacier
from dateutil.parser import parse
from django.core.management import CommandError

# upload to amazon glacier

def _get_vault_from_arn(arn, settings):
	g = Glacier(aws_access_key_id=settings.AWS_ACCESS_KEY_ID, aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY)
	for i in g.list_vaults():
		if arn == i.arn:
			return i
	else:
		raise CommandError('The specified vault could not be accessed.')

def upload(arn, output_file, settings):
	vault = _get_vault_from_arn(arn, settings)
	id = vault.upload_archive(output_file)
	
	# record backup internally
	# we don't need this record in order to restore from backup (obviously!)
	# but it makes pruning the backup set easier, and amazon reccomends it
	record = models.GlacierBackup.objects.create(glacier_id=id, date=datetime.now())
	record.save()

def reconcile(arn, settings):
	vault = _get_vault_from_arn(arn, settings)

	# check any inventory requests that have not been collected,
	# and if they are finished, collect them
	to_be_collected = models.GlacierInventory.objects.filter(collected_date=None)
	if len(to_be_collected) > 0:
		for record in to_be_collected:
			job = vault.get_job(record.inventory_id)
			if job.completed:
				print "Reconciling inventory", record.inventory_id
				_do_reconcile(job.get_output())
				record.collected_date = datetime.now()
				record.save()

	# if there are no collected inventories in the last 14 days,
	# and no requested inventories in the last 3,
	# request another inventory
	max_requested_date = datetime.now() - timedelta(days=3)
	max_collected_date = datetime.now() - timedelta(days=14)
	if models.GlacierInventory.objects.exclude(collected_date__lte=max_collected_date).exclude(requested_date__lte=max_requested_date).count() == 0:
		job_id = vault.retrieve_inventory()
		record = models.GlacierInventory.objects.create(inventory_id=job_id)
		record.save()

def _do_reconcile(inventory):
	for archive in inventory['ArchiveList']:
		id = archive['ArchiveId']
		creation_date = parse(archive['CreationDate'])
		if not models.GlacierBackup.objects.filter(glacier_id=id).exists():
			models.GlacierBackup.objects.create(glacier_id=id, date=creation_date).save()

def prune(arn, settings):
	vault = _get_vault_from_arn(arn, settings)
	keep_all_before = datetime.now() - timedelta(days=31)
	keep_daily_before = datetime.now() - timedelta(days=90)
	keep_weekly_before = datetime.now() - timedelta(days=365)
	oldest_date = models.GlacierBackup.objects.all().order_by('date')[0].date
	_do_delete(vault, 1, keep_all_before, keep_daily_before)
	_do_delete(vault, 30, keep_daily_before, keep_weekly_before)
	_do_delete(vault, 30, keep_weekly_before, oldest_date)

def _do_delete(vault, day_count, from_date, to_date):
	begin_date = from_date
	while begin_date >= to_date:
		end_date = begin_date - timedelta(days=day_count)
		if end_date < to_date:
			end_date = to_date
		qs = models.GlacierBackup.objects.filter(date__lt=end_date, date__gte=begin_date)
		# delete all but the most recent
		for record in qs[1:]:
			print "Deleting", record.glacier_id
			vault.delete(record.glacier_id)
			record.delete()



from django.db import models

class GlacierBackup (models.Model):
	glacier_id = models.CharField(max_length=138, unique=True, verbose_name="Glacier backup ID")
	date = models.DateTimeField()
	class Meta:
		verbose_name = "Glacier backup"
		verbose_name_plural = "Glacier backups"

class GlacierInventory (models.Model):
	inventory_id = models.CharField(max_length=92, unique=True, verbose_name="Glacier inventory ID")
	collected_date = models.DateTimeField(blank=True, null=True, default=None)
	requested_date = models.DateTimeField(auto_now_add=True)


import os
from django.core.management.base import CommandError
from tempfile import mkstemp
from subprocess import check_call
from shutil import copy
from django.conf import settings

BACKUP = 1
RESTORE = 2

MYSQL_BIN = getattr(settings, 'MYSQL_BIN', 'mysql')
MYSQLDUMP_BIN = getattr(settings, 'MYSQLDUMP_BIN', 'mysqldump')
PG_DUMP_BIN = getattr(settings, 'PG_DUMP_BIN', 'pg_dump')
PG_RESTORE_BIN = getattr(settings, 'PG_RESTORE_BIN', 'pg_restore')

def _database_dict_from_settings(settings):
	if hasattr(settings, 'DATABASES'):
		database_list = settings.DATABASES
	else:
		# database details are in the old format, so convert to the new one
		database_list = {
			'default': {
				'ENGINE': settings.DATABASE_ENGINE,
				'NAME': settings.DATABASE_NAME,
				'USER': settings.DATABASE_USER,
				'PASSWORD': settings.DATABASE_PASSWORD,
				'HOST': settings.DATABASE_HOST,
				'PORT': settings.DATABASE_PORT,
			}
		}
	return database_list

def backup_to(settings, dir, **kwargs):
	for name, database in _database_dict_from_settings(settings).iteritems():
		do(BACKUP, database, os.path.join(dir, name), **kwargs)

def restore_from(settings, dir, **kwargs):
	for name, database in _database_dict_from_settings(settings).iteritems():
		do(RESTORE, database, os.path.join(dir, name), **kwargs)

def do(action, database, f, **kwargs):
	engine = database['ENGINE']
	if 'mysql' in engine:
		__mysql(action, database, f)
	elif 'postgresql' in engine:
		if 'postgres_flags' not in kwargs:
			__postgresql(action, database, f)
		else:
			__postgresql(action, database, f, **kwargs)
	elif 'sqlite3' in engine:
		__sqlite(action, database, f)
	else:
		raise CommandError('{} in {} engine not implemented'.format('Backup' if action == BACKUP else 'Restore', engine))

def __sqlite(action, database, f):
	if action == BACKUP:
		copy(database['NAME'], f)
	elif action == RESTORE:
		copy(f, database['NAME'])

def __mysql(action, database, f):
	if action == BACKUP:
		command = [MYSQLDUMP_BIN]
	elif action == RESTORE:
		command = [MYSQL_BIN]

	if 'USER' in database:
		command += ["--user=%s" % database['USER']]
	if 'PASSWORD' in database:
		command += ["--password=%s" % database['PASSWORD']]
	if 'HOST' in database:
		command += ["--host=%s" % database['HOST']]
	if 'PORT' in database:
		command += ["--port=%s" % database['PORT']]
	command += [database['NAME']]

	if action == BACKUP:
		with open(f, 'w') as fo:
			check_call(command, stdout=fo)
	elif action == RESTORE:
		with open(f, 'r') as fo:
			check_call(command, stdin=fo)

def __postgresql(action, database, f, **kwargs):
	if action == BACKUP:
		command = [PG_DUMP_BIN, '--format=c']
	elif action == RESTORE:
		command = [PG_RESTORE_BIN, '-{}'.format(kwargs.get('postgres_flags', 'Oxc'))]

	if 'USER' in database and database['USER']:
		command.append("--username={}".format(database['USER']))
	if 'HOST' in database and database['HOST']:
		command.append("--host={}".format(database['HOST']))
	if 'PORT' in database and database['PORT']:
		command.append("--port={}".format(database['PORT']))
	if 'NAME' in database and database['NAME']:
		if action == RESTORE:
			command.append('--dbname={}'.format(database['NAME']))
		else:
			command.append(database['NAME'])

	if 'PASSWORD' in database and database['PASSWORD']:
		# create a pgpass file that always returns the same password, as a secure temp file
		password_fd, password_path = mkstemp()
		password_file = os.fdopen(password_fd, 'w')
		password_file.write('*:*:*:*:{}'.format(database['PASSWORD']))
		password_file.close()
		os.environ['PGPASSFILE'] = password_path
	else:
		command.append('-w')

	if action == BACKUP:
		with open(f, 'w') as fo:
			check_call(command, stdout=fo)
	elif action == RESTORE:
		with open(f, 'r') as fo:
			check_call(command, stdin=fo)

	# clean up
	if 'PASSWORD' in database and database['PASSWORD']:
		os.remove(password_path)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding field 'GlacierInventory.requested_date'
        db.add_column('icybackup_glacierinventory', 'requested_date',
                      self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, default=datetime.datetime(2013, 1, 9, 0, 0), blank=True),
                      keep_default=False)


    def backwards(self, orm):
        # Deleting field 'GlacierInventory.requested_date'
        db.delete_column('icybackup_glacierinventory', 'requested_date')


    models = {
        'icybackup.glacierbackup': {
            'Meta': {'object_name': 'GlacierBackup'},
            'date': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'glacier_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '138'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'})
        },
        'icybackup.glacierinventory': {
            'Meta': {'object_name': 'GlacierInventory'},
            'collected_date': ('django.db.models.fields.DateTimeField', [], {'default': 'None', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'inventory_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '92'}),
            'requested_date': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['icybackup']


import os, sys, time
from optparse import make_option
from tempfile import mkdtemp, NamedTemporaryFile
import tarfile
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from ...components import db, glacier, lib

# Based on: http://code.google.com/p/django-backup/
# Based on: http://www.djangosnippets.org/snippets/823/
# Based on: http://www.yashh.com/blog/2008/sep/05/django-database-backup-view/
class Command(BaseCommand):
	option_list = BaseCommand.option_list + (
		make_option('-o', '--output', default=None, dest='output',
			help='Write backup to file'),
		make_option('-d', '--outdir', default=None, dest='outdir',
			help='Write backup to timestamped file in a directory'),
		make_option('-g', '--glacier', default=None, dest='glacier',
			help='Upload backup to the Amazon Glacier vault with the given ARN'),
		make_option('-O', '--stdout', action='store_true', dest='stdout',
			help='Output backup tarball to standard output'),
		make_option('--extra', '-e', action='append', default=[], dest='extras',
			help='Include extra directories or files in the backup tarball'),
	)
	help = "Back up a Django installation (database and media directory)."

	def handle(self, *args, **options):
		extras = options.get('extras')

		output_file = options.get('output')
		output_dir = options.get('outdir')
		glacier_vault = options.get('glacier')
		output_to_stdout = options.get('stdout')
		output_file_temporary = False

		# glacier backups go to a temporary file
		if glacier_vault is not None or output_to_stdout:
			output_file_temporary = True
			output_file_obj = NamedTemporaryFile(delete=False)
			output_file_obj.close() # we'll open it later
			output_file = output_file_obj.name

		if output_file is None:
			if output_dir is None:
				raise CommandError('You must specify an output file')
			else:
				output_file = os.path.join(output_dir, '{}.tgz'.format(_time()))

		media_root = settings.MEDIA_ROOT

		# Create a temporary directory to perform our backup in
		backup_root = mkdtemp()
		database_root = os.path.join(backup_root, 'databases')
		os.mkdir(database_root)

		# Back up databases
		db.backup_to(settings, database_root)

		# create backup gzipped tarball
		with tarfile.open(output_file, 'w:gz') as tf:
			tf.add(database_root, arcname='backup/databases')
			tf.add(media_root, arcname='backup/media')
			if len(extras) > 0:
				extras_mf = NamedTemporaryFile(delete=False)
				for count, extra in enumerate(extras):
					tf.add(extra, arcname='backup/extras/{}'.format(count))
					extras_mf.write('{},{}\n'.format(count, extra.replace(',','\\,')))
				extras_mf.close()
				tf.add(extras_mf.name, arcname='backup/extras/manifest')
				os.unlink(extras_mf.name)

		# upload to glacier
		if glacier_vault is not None:
			glacier.upload(glacier_vault, output_file, settings)
			glacier.reconcile(glacier_vault, settings)
			glacier.prune(glacier_vault, settings)

		# output to stdout
		if output_to_stdout:
			with open(output_file, 'r') as f:
				sys.stdout.write(f.read())

		# clean up
		lib.rm_rf(backup_root)
		if output_file_temporary:
			os.unlink(output_file)

def _time():
	return time.strftime('%Y%m%d-%H%M%S')
